#include "Widget.h"
#include "ui_Widget.h"

#include <QGraphicsPathItem>
#include <QDebug>
#include <QFile>

#define BUG_BEHAVIOUR 1
#define WORKAROUND 0

Widget::Widget(QWidget* p_parent/* = 0*/)
    : QWidget(p_parent)
    , m_ui(new Ui::Widget)
    , m_aroundItem(0)
    , m_otherItems()
    , m_pathItemNames()
{
    m_ui->setupUi(this);

    bool connected = true;
    connected &= static_cast<bool>(connect(m_ui->m_exitButton, SIGNAL(clicked(bool)), qApp, SLOT(quit())));
    Q_ASSERT(connected);

    BuildScene();
    GetCollidings();
}

Widget::~Widget()
{
    delete m_ui;
}


void Widget::BuildScene()
{
    m_ui->m_graphicsView->setScene(new QGraphicsScene());

    {
        // Add painter path that is around others
        QPainterPath path;
        path.moveTo(60, 20);
        path.lineTo(20, 30);
        path.lineTo(30, 60);
        path.lineTo(60, 90);
        path.lineTo(90, 50);
        path.lineTo(60, 20);

        m_aroundItem = new QGraphicsPathItem(path);
        QPen pen(Qt::blue);
        pen.setCosmetic(true);
        m_aroundItem->setPen(pen);
        m_pathItemNames[m_aroundItem] = "blue";

        m_ui->m_graphicsView->scene()->addItem(m_aroundItem);

        // Add around item bounding rect
        pen.setStyle(Qt::DashLine);
        m_ui->m_graphicsView->scene()->addRect(path.boundingRect(), pen);

        // This is the biggest element, set scene rect and zoom on it
        const qreal scale = qMax(m_ui->m_graphicsView->width() * 2 / m_ui->m_graphicsView->sceneRect().width(),
                                 m_ui->m_graphicsView->height() * 2 / m_ui->m_graphicsView->sceneRect().height());
        m_ui->m_graphicsView->scale(scale, scale);
        m_ui->m_graphicsView->centerOn(path.boundingRect().center());
    }

    {
        // Add painter path to item inside "around" item that is actually seen inside
        QPainterPath path;
        path.moveTo(50, 40);
        path.lineTo(40, 50);
        path.lineTo(50, 50);
        path.lineTo(50, 40);

        QGraphicsPathItem* pathItem = new QGraphicsPathItem(path);
        QPen pen(Qt::green);
        pen.setCosmetic(true);
        pathItem->setPen(pen);
        m_pathItemNames[pathItem] = "green";

        m_ui->m_graphicsView->scene()->addItem(pathItem);
        m_otherItems << pathItem;
    }

    {
        // Add painter path to item inside "around" item that is actually seen inside
        QPainterPath path;
#if BUG_BEHAVIOUR
        // Buggy behaviour (red shape not found inside "around" item)
        path.moveTo(50, 68);
        path.lineTo(40, 68);
        path.lineTo(50, 78);
        path.lineTo(50, 68);
#else
        // Non-buggy behaviour (red shape found inside "around" item after having moved it more inside)
        path.moveTo(50, 60);
        path.lineTo(40, 60);
        path.lineTo(50, 70);
        path.lineTo(50, 60);
#endif

        QGraphicsPathItem* pathItem = new QGraphicsPathItem(path);
        QPen pen(Qt::red);
        pen.setCosmetic(true);
        pathItem->setPen(pen);
        m_pathItemNames[pathItem] = "red";

        m_ui->m_graphicsView->scene()->addItem(pathItem);
        m_otherItems << pathItem;
    }

    {
        // Add painter path to item inside "around" item that is actually seen inside
        QPainterPath path;
        path.moveTo(35, 70);
        path.lineTo(25, 70);
        path.lineTo(25, 80);
        path.lineTo(35, 80);
        path.lineTo(35, 70);

        QGraphicsPathItem* pathItem = new QGraphicsPathItem(path);
        QPen pen(Qt::darkRed);
        pen.setCosmetic(true);
        pathItem->setPen(pen);
        m_pathItemNames[pathItem] = "darkRed";

        m_ui->m_graphicsView->scene()->addItem(pathItem);
        m_otherItems << pathItem;
    }

    Q_FOREACH(QGraphicsPathItem* pathItem, m_otherItems)
    {
        if (m_aroundItem->path().contains(pathItem->path()))
        {
            qDebug().nospace().noquote() << "m_aroundItem contains [" << m_pathItemNames[pathItem] << "] item";
        }
        else
        {
            qDebug().nospace().noquote() << "m_aroundItem does not contain [" << m_pathItemNames[pathItem] << "] item";
        }
    }
}

void Widget::GetCollidings()
{
    {
        // Scene items method
        QList<QGraphicsItem*> elementsInsideAround;
#if WORKAROUND
        const QPainterPath aroundItemShape(m_aroundItem->shape());
        // Get elements under m_aroundItem bounding box, we will get more than we want but...
        QList<QGraphicsItem*> elementsInsideAroundBoundingBox = m_ui->m_graphicsView->scene()->items(aroundItemShape.boundingRect(), Qt::ContainsItemShape);
        Q_FOREACH(QGraphicsItem* item, elementsInsideAroundBoundingBox)
        {
            if (aroundItemShape.contains(item->shape()))
            {
                // ... here we only add elements that are actually under m_aroundItem
                elementsInsideAround << item;
            }
        }

        qDebug().nospace().noquote() << "Scene items method workaround (under bounding box): found [" << elementsInsideAroundBoundingBox.size() << "] item(s)";
        Q_FOREACH(QGraphicsItem* item, elementsInsideAroundBoundingBox)
        {
            QGraphicsPathItem* pathItem = static_cast<QGraphicsPathItem*>(item);
            qDebug().nospace().noquote() << "    - " << m_pathItemNames[pathItem];
        }
        qDebug().nospace().noquote() << "Scene items method workaround (under shape): found [" << elementsInsideAround.size() << "] item(s)";
#else
        elementsInsideAround = m_ui->m_graphicsView->scene()->items(m_aroundItem->shape(), Qt::ContainsItemShape);
        qDebug().nospace().noquote() << "Scene items method: found [" << elementsInsideAround.size() << "] item(s)";
#endif
        Q_FOREACH(QGraphicsItem* item, elementsInsideAround)
        {
            QGraphicsPathItem* pathItem = static_cast<QGraphicsPathItem*>(item);
            qDebug().nospace().noquote() << "    - " << m_pathItemNames[pathItem];
        }
    }

    {
        // Scene collidingItems method
        QList<QGraphicsItem*> elementsInsideAround = m_ui->m_graphicsView->scene()->collidingItems(m_aroundItem, Qt::ContainsItemShape);
        qDebug().nospace().noquote() << "Scene collidingItems method: found [" << elementsInsideAround.size() << "] item(s)";
        Q_FOREACH(QGraphicsItem* item, elementsInsideAround)
        {
            QGraphicsPathItem* pathItem = static_cast<QGraphicsPathItem*>(item);
            qDebug().nospace().noquote() << "    - " << m_pathItemNames[pathItem];
        }
    }

    {
        // Item collidingItems method
        QList<QGraphicsItem*> elementsInsideAround = m_aroundItem->collidingItems(Qt::ContainsItemShape);
        qDebug().nospace().noquote() << "Item collidingItems method: found [" << elementsInsideAround.size() << "] item(s)";
        Q_FOREACH(QGraphicsItem* item, elementsInsideAround)
        {
            QGraphicsPathItem* pathItem = static_cast<QGraphicsPathItem*>(item);
            qDebug().nospace().noquote() << "    - " << m_pathItemNames[pathItem];
        }
    }

    {
        // Item collidesWithItem method
        QList<QGraphicsPathItem*> elementsInsideAround;
        Q_FOREACH(QGraphicsPathItem* pathItem, m_otherItems)
        {
            if (m_aroundItem->collidesWithItem(pathItem, Qt::ContainsItemShape))
            {
                elementsInsideAround << pathItem;
            }
        }
        qDebug().nospace().noquote() << "Item collidesWithItem method: found [" << elementsInsideAround.size() << "] item(s)";
        Q_FOREACH(QGraphicsPathItem* pathItem, elementsInsideAround)
        {
            qDebug().nospace().noquote() << "    - " << m_pathItemNames[pathItem];
        }
    }

    {
        // Item collidesWithPath method
        QList<QGraphicsPathItem*> elementsInsideAround;
        Q_FOREACH(QGraphicsPathItem* pathItem, m_otherItems)
        {
            if (m_aroundItem->collidesWithPath(pathItem->shape(), Qt::ContainsItemShape))
            {
                elementsInsideAround << pathItem;
            }
        }
        qDebug().nospace().noquote() << "Item collidesWithPath method: found [" << elementsInsideAround.size() << "] item(s)";
        Q_FOREACH(QGraphicsPathItem* pathItem, elementsInsideAround)
        {
            qDebug().nospace().noquote() << "    - " << m_pathItemNames[pathItem];
        }
    }
}
